package at.ac.uibk.gitsearch.service;

// for static metamodels
import at.ac.uibk.gitsearch.domain.ReviewHistoryEntry_;
import at.ac.uibk.gitsearch.domain.ReviewRating;
import at.ac.uibk.gitsearch.domain.ReviewRating_;
import at.ac.uibk.gitsearch.domain.Review_;
import at.ac.uibk.gitsearch.domain.User_;
import at.ac.uibk.gitsearch.repository.ReviewRatingRepository;
import at.ac.uibk.gitsearch.service.criteria.ReviewRatingCriteria;
import at.ac.uibk.gitsearch.service.dto.ReviewRatingDTO;
import at.ac.uibk.gitsearch.service.mapper.ReviewRatingMapper;
import java.util.List;
import javax.persistence.criteria.JoinType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import tech.jhipster.service.QueryService;

/**
 * Service for executing complex queries for {@link ReviewRating} entities in the database.
 * The main input is a {@link ReviewRatingCriteria} which gets converted to {@link Specification},
 * in a way that all the filters must apply.
 * It returns a {@link List} of {@link ReviewRatingDTO} or a {@link Page} of {@link ReviewRatingDTO} which fulfills the criteria.
 */
@Service
@Transactional(readOnly = true)
@SuppressWarnings({ "PMD.CognitiveComplexity" })
public class ReviewRatingQueryService extends QueryService<ReviewRating> {

    private final Logger log = LoggerFactory.getLogger(ReviewRatingQueryService.class);
    private final ReviewRatingRepository reviewRatingRepository;

    private final ReviewRatingMapper reviewRatingMapper;

    public ReviewRatingQueryService(ReviewRatingRepository reviewRatingRepository, ReviewRatingMapper reviewRatingMapper) {
        super();
        this.reviewRatingRepository = reviewRatingRepository;
        this.reviewRatingMapper = reviewRatingMapper;
    }

    /**
     * Return a {@link List} of {@link ReviewRatingDTO} which matches the criteria from the database.
     * @param criteria The object which holds all the filters, which the entities should match.
     * @return the matching entities.
     */
    @Transactional(readOnly = true)
    public List<ReviewRatingDTO> findByCriteria(ReviewRatingCriteria criteria) {
        log.debug("find by criteria : {}", criteria);
        final Specification<ReviewRating> specification = createSpecification(criteria);
        return reviewRatingMapper.toDto(reviewRatingRepository.findAll(specification));
    }

    /**
     * Return a {@link Page} of {@link ReviewRatingDTO} which matches the criteria from the database.
     * @param criteria The object which holds all the filters, which the entities should match.
     * @param page The page, which should be returned.
     * @return the matching entities.
     */
    @Transactional(readOnly = true)
    public Page<ReviewRatingDTO> findByCriteria(ReviewRatingCriteria criteria, Pageable page) {
        log.debug("find by criteria : {}, page: {}", criteria, page);
        final Specification<ReviewRating> specification = createSpecification(criteria);
        return reviewRatingRepository.findAll(specification, page).map(reviewRatingMapper::toDto);
    }

    /**
     * Return the number of matching entities in the database.
     * @param criteria The object which holds all the filters, which the entities should match.
     * @return the number of matching entities.
     */
    @Transactional(readOnly = true)
    public long countByCriteria(ReviewRatingCriteria criteria) {
        log.debug("count by criteria : {}", criteria);
        final Specification<ReviewRating> specification = createSpecification(criteria);
        return reviewRatingRepository.count(specification);
    }

    /**
     * Function to convert {@link ReviewRatingCriteria} to a {@link Specification}
     * @param criteria The object which holds all the filters, which the entities should match.
     * @return the matching {@link Specification} of the entity.
     */
    protected Specification<ReviewRating> createSpecification(ReviewRatingCriteria criteria) {
        Specification<ReviewRating> specification = Specification.where(null);
        if (criteria != null) {
            // This has to be called first, because the distinct method returns null
            if (criteria.getDistinct() != null) {
                specification = specification.and(distinct(criteria.getDistinct()));
            }
            if (criteria.getId() != null) {
                specification = specification.and(buildRangeSpecification(criteria.getId(), ReviewRating_.id));
            }
            if (criteria.getStatus() != null) {
                specification = specification.and(buildSpecification(criteria.getStatus(), ReviewRating_.status));
            }
            if (criteria.getReviewId() != null) {
                specification =
                    specification.and(
                        buildSpecification(criteria.getReviewId(), root -> root.join(ReviewRating_.review, JoinType.LEFT).get(Review_.id))
                    );
            }
            if (criteria.getUserId() != null) {
                specification =
                    specification.and(
                        buildSpecification(criteria.getUserId(), root -> root.join(ReviewRating_.user, JoinType.LEFT).get(User_.id))
                    );
            }
            if (criteria.getHistoryId() != null) {
                specification =
                    specification.and(
                        buildSpecification(
                            criteria.getHistoryId(),
                            root -> root.join(ReviewRating_.histories, JoinType.LEFT).get(ReviewHistoryEntry_.id)
                        )
                    );
            }
        }
        return specification;
    }
}
