package at.ac.uibk.gitsearch.web.rest;

import at.ac.uibk.gitsearch.es.model.ArtemisExerciseInfo;
import at.ac.uibk.gitsearch.security.SecurityUtils;
import at.ac.uibk.gitsearch.service.ArtemisImportError;
import at.ac.uibk.gitsearch.service.ExerciseImportService;
import at.ac.uibk.gitsearch.service.GitlabService;
import at.ac.uibk.gitsearch.service.SearchService;
import at.ac.uibk.gitsearch.service.SearchService.ExtractionDepth;
import at.ac.uibk.gitsearch.service.StatisticsService;
import at.ac.uibk.gitsearch.service.dto.StatisticsDTO;
import at.ac.uibk.gitsearch.web.rest.utils.RestUtils;
import at.ac.uibk.gitsearch.web.util.HeaderUtil;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.nio.file.Files;
import java.text.ParseException;
import java.util.Optional;
import javax.servlet.http.HttpServletRequest;
import javax.ws.rs.core.Response;
import org.apache.commons.lang3.StringUtils;
import org.codeability.sharing.plugins.api.search.SearchResultDTO;
import org.codeability.sharing.plugins.api.search.util.ExerciseId;
import org.eclipse.jgit.api.errors.GitAPIException;
import org.gitlab4j.api.GitLabApiException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.core.io.InputStreamResource;
import org.springframework.core.io.Resource;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RestController;
import org.springframework.web.servlet.HandlerMapping;

/**
 * this resource should give access to various aspects of an exercise or course
 * @author Michael Breu
 *
 */
@RestController
@RequestMapping("/api")
public class ExerciseResource {

    private static final String THERE_WAS_AN_ERROR_FINDING_YOUR_EXERCISE_FORMAT = "There was an error finding your exercise %s.";

    private static final String NOT_FOUND = "not found";

    private static final String ENTITY_NAME = "ExerciseResource";

    private static final String REPOSITORYFILE = "RepositoryFile";

    private final Logger log = LoggerFactory.getLogger(ExerciseResource.class);

    @Value("${application.registeredConnectorsCallBackURL}")
    String baseApiUrl;

    @Autowired
    @SuppressWarnings({ "PMD.ImmutableField", "PMD.AvoidDuplicateLiterals" })
    private GitlabService gitLabService;

    @Autowired
    @SuppressWarnings({ "PMD.ImmutableField", "PMD.AvoidDuplicateLiterals" })
    private ExerciseImportService exerciseImportService;

    @Autowired
    @SuppressWarnings({ "PMD.ImmutableField", "PMD.AvoidDuplicateLiterals" })
    private SearchService searchService;

    @Autowired
    @SuppressWarnings({ "PMD.ImmutableField", "PMD.AvoidDuplicateLiterals" })
    private StatisticsService statisticsService;

    @Value("${jhipster.clientApp.name}")
    @SuppressWarnings({ "PMD.ImmutableField", "PMD.AvoidDuplicateLiterals" })
    private String applicationName;

    /**
     * GET some file of an exercise.
     *
     * @param filePath the file path
     * @return the ResponseEntity with status 200 (OK) and with body the exercise file, or with status 404 (Not Found)
     */
    @GetMapping("/exerciseFile/**")
    @SuppressWarnings("PMD.CloseResource")
    public ResponseEntity<Resource> getRepositoryFile(HttpServletRequest request, @RequestParam("filePath") String filePath)
        throws IOException {
        String pattern = (String) request.getAttribute(HandlerMapping.BEST_MATCHING_PATTERN_ATTRIBUTE);

        String exerciseId = RestUtils.decodeURL(request, pattern);

        ExerciseId parsedId;
        try {
            parsedId = ExerciseId.fromString(exerciseId);
        } catch (ParseException e) {
            return ResponseEntity
                .notFound()
                .headers(
                    HeaderUtil.createFailureAlert(
                        applicationName,
                        true,
                        REPOSITORYFILE,
                        NOT_FOUND,
                        String.format(THERE_WAS_AN_ERROR_FINDING_YOUR_EXERCISE_FORMAT, exerciseId)
                    )
                )
                .build();
        }

        InputStream repositoryStream;
        try {
            repositoryStream = gitLabService.getRepositoryFile(parsedId, filePath);
            if (repositoryStream == null) {
                return ResponseEntity
                    .notFound()
                    .headers(
                        HeaderUtil.createFailureAlert(
                            applicationName,
                            true,
                            REPOSITORYFILE,
                            NOT_FOUND,
                            "The requested resource file does not exist or is not available."
                        )
                    )
                    .build();
            }
        } catch (IOException | GitLabApiException e) {
            log.warn("Cannot read file for exercise {} and path {}: {}", exerciseId, filePath, e.getMessage());
            return ResponseEntity
                .notFound()
                .headers(
                    HeaderUtil.createFailureAlert(
                        applicationName,
                        true,
                        REPOSITORYFILE,
                        NOT_FOUND,
                        "There was an error on the server and the resource file could not be created."
                    )
                )
                .build();
        }

        InputStreamResource resource = new InputStreamResource(repositoryStream);

        return ResponseEntity.ok().contentType(MediaType.APPLICATION_OCTET_STREAM).header("filename", filePath).body(resource);
    }

    /**
     * GET the exercise by Id
     *
     * @param request the complete request, in order to parse the exerciseId
     * @return the ResponseEntity with status 200 (OK) and with the searchresult, or with status 404 (Not Found)
     */
    @GetMapping("/exercise/**")
    public ResponseEntity<SearchResultDTO> getExerciseById(HttpServletRequest request) {
        String pattern = (String) request.getAttribute(HandlerMapping.BEST_MATCHING_PATTERN_ATTRIBUTE);

        String exerciseId = RestUtils.decodeURL(request, pattern);

        ExerciseId parsedId;
        try {
            parsedId = ExerciseId.fromString(exerciseId);
        } catch (ParseException e) {
            log.warn("Cannot parse exercise {}", exerciseId);

            return ResponseEntity
                .notFound()
                .headers(
                    HeaderUtil.createFailureAlert(
                        applicationName,
                        true,
                        REPOSITORYFILE,
                        NOT_FOUND,
                        "There was an error with your requested id " + exerciseId + "."
                    )
                )
                .build();
        }
        log.debug("Loading exercise {}", parsedId);
        final Optional<SearchResultDTO> result = searchService.findExerciseById(parsedId);

        if (result.isPresent()) {
            return ResponseEntity.ok().contentType(MediaType.APPLICATION_JSON).body(result.get());
        } else {
            return ResponseEntity
                .notFound()
                .headers(
                    HeaderUtil.createFailureAlert(
                        applicationName,
                        true,
                        REPOSITORYFILE,
                        NOT_FOUND,
                        String.format(THERE_WAS_AN_ERROR_FINDING_YOUR_EXERCISE_FORMAT, exerciseId)
                    )
                )
                .build();
        }
    }

    /**
     * POST /exerciseFiles/:exerciseId: sends all repositories and details of the programming exercise as zip.
     * However excepted files/folders are omitted, if exercise is publicVisibility
     *
     * @param exerciseId the id of the exercise to get the repos from
     * @param recursion the recursion depth for the children
     * ResponseEntity with status
     * @throws IOException if something during the zip process went wrong
     */
    @PostMapping("/exerciseFiles/**")
    public ResponseEntity<Resource> exportProgrammingExercise(HttpServletRequest request) throws IOException {
        String pattern = (String) request.getAttribute(HandlerMapping.BEST_MATCHING_PATTERN_ATTRIBUTE);
        String recursionParam = request.getParameter("recursion");
        ExtractionDepth recursion = ExtractionDepth.JUST_PROJECT;
        if (!StringUtils.isEmpty(recursionParam)) {
            try {
                recursion = ExtractionDepth.valueOf(recursionParam);
            } catch (IllegalArgumentException e) {
                log.warn("Cannog parse {}", recursionParam);
                recursion = ExtractionDepth.JUST_PROJECT;
            }
        }
        String exerciseId = RestUtils.decodeURL(request, pattern);

        ExerciseId parsedId;
        try {
            parsedId = ExerciseId.fromString(exerciseId);
        } catch (ParseException e) {
            return ResponseEntity
                .notFound()
                .headers(
                    HeaderUtil.createFailureAlert(
                        applicationName,
                        true,
                        REPOSITORYFILE,
                        NOT_FOUND,
                        String.format(THERE_WAS_AN_ERROR_FINDING_YOUR_EXERCISE_FORMAT, exerciseId)
                    )
                )
                .build();
        }

        File zipFile = searchService.exportExercise(parsedId, recursion);
        if (zipFile == null) {
            return ResponseEntity
                .badRequest()
                .headers(
                    HeaderUtil.createFailureAlert(
                        applicationName,
                        true,
                        ENTITY_NAME,
                        "internalServerError",
                        "There was an error on the server and the zip file could not be created."
                    )
                )
                .body(null);
        }

        /*
         * Customized FileInputStream to delete and therefore clean up the returned
         * files
         */
        class NewFileInputStream extends FileInputStream {

            File file;

            public NewFileInputStream(File file) throws FileNotFoundException {
                super(file);
                this.file = file;
            }

            @Override
            public void close() throws IOException {
                super.close();
                try {
                    Files.delete(file.toPath());
                } catch (IOException e) {
                    log.info("Cannot delete {}", file.getAbsoluteFile(), e);
                }
            }
        }
        InputStreamResource resource = new InputStreamResource(new NewFileInputStream(zipFile));

        log.debug("REST request to get Statistics for ExerciseID : {}", exerciseId);
        Optional<StatisticsDTO> statisticsDTO = statisticsService.findOneByExerciseID(exerciseId);
        if (statisticsDTO.isPresent()) {
            StatisticsDTO newStats = statisticsDTO.get();
            newStats.setDownloads(newStats.getDownloads() + 1);
            statisticsService.save(newStats);
            log.debug("REST increased number of downloads for ExerciseID : {}", exerciseId);
        }
        if (!statisticsDTO.isPresent()) {
            StatisticsDTO newStats = new StatisticsDTO();
            newStats.setDownloads(1);
            newStats.setViews(1);
            newStats.setExerciseID(exerciseId);
            newStats.setBadgeRewarded(false);
            statisticsService.save(newStats);
            log.debug("Created new statistics entry for exerciseID: {}", exerciseId);
        }

        return ResponseEntity
            .ok()
            .contentLength(zipFile.length())
            .contentType(MediaType.APPLICATION_OCTET_STREAM)
            .header("filename", "exercise" + exerciseId + ".zip")
            .body(resource);
    }

    /**
     * GET /exercise/import : Endpoint listening for request to import exercises.
     * Ex. ArTEMiS calls this endpoint and provides an URL where the exercise can be retrieved
     * @param exerciseUrl URL where the exercise can be retrieved by GitSearch
     * @param apiKey Key used to validate the request
     * @param request Used to retrieve the baseUrl of the request
     * @return URI to the client page containing the exercise import form
     * @throws ArtemisImportError
     * @throws MalformedURLException
     * @throws URISyntaxException
     */
    @GetMapping("/exercise/import")
    public ResponseEntity<URI> importExerciseFromArtemis(
        @RequestParam("exerciseUrl") String exerciseUrl,
        @RequestParam("apiKey") Optional<String> apiKey,
        HttpServletRequest request
    ) throws ArtemisImportError, MalformedURLException, URISyntaxException {
        if (apiKey.isPresent() && exerciseImportService.validate(apiKey.get())) {
            exerciseImportService.importExerciseFromArtemis(exerciseUrl);
        } else {
            return ResponseEntity.status(401).body(null);
        }

        String baseUrl = getBaseUrl(request);
        return ResponseEntity.ok(new URL(baseUrl + "/import/" + exerciseImportService.getTokenFromUrl(exerciseUrl)).toURI());
    }

    /**
     * GET /exercise/imported-exercise-info/{exerciseToken} : Used to retrieve the exercise's metadata
     * of a given exerciseToken
     * @param exerciseToken of the exercise to retrieve
     * @return Object containing all metadata of an ArTEMiS exercise
     * @throws IOException
     */
    @GetMapping("/exercise/imported-exercise-info/{exerciseToken}")
    @SuppressWarnings({ "PMD.AvoidCatchingGenericException", "PMD.AvoidCatchingNPE" })
    public ResponseEntity<ArtemisExerciseInfo> getImportedExerciseInfo(@PathVariable("exerciseToken") String exerciseToken)
        throws IOException {
        if (SecurityUtils.isAuthenticated()) {
            try {
                return ResponseEntity.ok(exerciseImportService.getArtemisExerciseInfo(exerciseToken));
            } catch (NullPointerException npe) {
                return ResponseEntity.notFound().build();
            }
        } else {
            return ResponseEntity.status(HttpStatus.UNAUTHORIZED).body(null);
        }
    }

    /**
     * POST /exercise/import-exercise/{exerciseToken}/{gitlabGroupId} : Endpoint requesting the actual
     * import to GitSearch/GitLab.
     * @param exerciseInfo Exercise's metadata object
     * @param exerciseToken of the exercise to import
     * @param gitlabGroupId of the Gitlab group in which the exercise repository should be created
     * @return Response with redirect if the import was successful
     * @throws GitLabApiException
     * @throws GitAPIException
     * @throws IOException
     */
    @PostMapping("/exercise/import-exercise/{exerciseToken}/{gitlabGroupId}")
    public Response importExercise(
        @RequestBody ArtemisExerciseInfo exerciseInfo,
        @PathVariable("exerciseToken") String exerciseToken,
        @PathVariable("gitlabGroupId") Long gitlabGroupId
    ) throws GitLabApiException, IOException, ArtemisImportError {
        if (SecurityUtils.isAuthenticated()) {
            exerciseImportService.importExercise(exerciseInfo, exerciseToken, gitlabGroupId);
            return Response.ok().build();
        } else {
            return Response.status(401).build();
        }
    }

    @PostMapping("/exercise/validate-exercise-info")
    public ResponseEntity<String> validateExerciseInfo(@RequestBody ArtemisExerciseInfo exerciseInfo) {
        try {
            exerciseImportService.validateExerciseInfo(exerciseInfo);
            return ResponseEntity.ok().build();
        } catch (ArtemisImportError e) {
            return ResponseEntity.badRequest().body(e.getMessage());
        }
    }

    /**
     * Utility method used to retrieve the Base URL from a HTTP request
     * @param request the request to extract the Url from
     * @return String of the extracted Base URL
     */
    private static String getBaseUrl(HttpServletRequest request) {
        String scheme = request.getScheme() + "://";
        String serverName = request.getServerName();
        String serverPort = (request.getServerPort() == 80) ? "" : ":" + request.getServerPort();
        String contextPath = request.getContextPath();
        return scheme + serverName + serverPort + contextPath;
    }
}
