package at.ac.uibk.gitsearch.service.vocabulary;

import static org.assertj.core.api.Assertions.assertThat;
import static org.hamcrest.CoreMatchers.hasItem;
import static org.hamcrest.CoreMatchers.is;
import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.containsInAnyOrder;
import static org.hamcrest.Matchers.hasItemInArray;
import static org.hamcrest.Matchers.hasKey;
import static org.hamcrest.Matchers.hasProperty;

import at.ac.uibk.gitsearch.GitsearchApp;
import at.ac.uibk.gitsearch.domain.vocabulary.VocabularyItem;
import at.ac.uibk.gitsearch.domain.vocabulary.VocabularyItem.LanguageItem;
import at.ac.uibk.gitsearch.repository.vocabulary.VocabularyRepository;
import at.ac.uibk.gitsearch.service.gitlab_events.GitlabEventService.ExtUserProvidedMetadataDTO;
import at.ac.uibk.gitsearch.service.vocabulary.VocabularyService.AlternativesProposer;
import at.ac.uibk.gitsearch.service.vocabulary.VocabularyService.ValidationResultDTO;
import at.ac.uibk.gitsearch.service.vocabulary.VocabularyService.VocabularyServiceConfig;
import at.ac.uibk.gitsearch.service.vocabulary.VocabularyService.VocabularyServiceConfig.VocabularySetting;
import com.fasterxml.jackson.core.exc.StreamReadException;
import com.fasterxml.jackson.databind.DatabindException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.dataformat.yaml.YAMLFactory;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Stream;
import org.assertj.core.data.Offset;
import org.codeability.sharing.plugins.api.search.UserProvidedMetadataDTO;
import org.junit.Assert;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.TestInstance;
import org.junit.jupiter.api.TestInstance.Lifecycle;
import org.junit.jupiter.api.Timeout;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.CsvSource;
import org.junit.jupiter.params.provider.MethodSource;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.security.test.context.support.WithMockUser;

@SpringBootTest(classes = GitsearchApp.class)
@WithMockUser(value = "test", authorities = "sharing")
@TestInstance(Lifecycle.PER_CLASS)
class VocabularyServiceTest {

    @Autowired
    private VocabularyService vocabularyService;

    @Autowired
    private VocabularyRepository vocabularyRepository;

    @Test
    @Timeout(60000)
    void initTest() {
        final Map<String, URI> vocabularyURLMap = vocabularyService.getVocabularyURLMap();
        assertThat(vocabularyURLMap, hasKey("Keyword"));

        vocabularyURLMap
            .entrySet()
            .stream()
            .forEach(entry -> System.out.println("Type: " + entry.getKey() + " Url: " + entry.getValue().toASCIIString()));
    }

    @Test
    @Timeout(60000)
    void keywordLoadTest() {
        final VocabularyItem[] keywordItems = vocabularyService.getKeywordItems();
        assertThat(keywordItems, hasItemInArray(hasProperty("languageItem", hasItemInArray(hasProperty("title", is("Variables"))))));
    }

    @Test
    @Timeout(60000)
    void programmingLanguageItemsTest() {
        final VocabularyItem[] keywordItems = vocabularyService.getProgrammingLanguageItems();
        assertThat(keywordItems, hasItemInArray(hasProperty("languageItem", hasItemInArray(hasProperty("title", is("Java"))))));
    }

    @Test
    @Timeout(60000)
    void difficultyItemsTest() {
        final VocabularyItem[] keywordItems = vocabularyService.getDifficultyItems();
        assertThat(keywordItems, hasItemInArray(hasProperty("languageItem", hasItemInArray(hasProperty("title", is("easy"))))));
        assertThat(keywordItems, hasItemInArray(hasProperty("languageItem", hasItemInArray(hasProperty("title", is("leicht"))))));
    }

    private Stream<Arguments> getAllVocabularies() {
        final Map<String, URI> vocabularyURLMap = vocabularyService.getVocabularyURLMap();
        return vocabularyURLMap
            .entrySet()
            .stream()
            .filter(entry -> !(entry.getKey().equals("Description") || entry.getKey().equals("Status") || entry.getKey().equals("Title")))
            .map(entry -> Arguments.of(entry.getKey(), entry.getValue()));
    }

    @ParameterizedTest
    @MethodSource("getAllVocabularies")
    @Timeout(60000)
    void checkAllURIs(String key, URI target) {
        VocabularyItem[] items = vocabularyRepository.getVocabularyItemsFor(key);
        if (!key.equals("Creators") && !key.equals("Contributors") && !key.equals("Publisher")) {
            Assert.assertTrue(
                "vocabulary for " + key + "[" + target.toASCIIString() + "] should contain at least one item",
                items.length > 0
            );
        }
    }

    @Test
    @Timeout(60000)
    void testProposeAlternative() {
        VocabularyItem[] items = vocabularyRepository.getVocabularyItemsFor("License");

        final Optional<String> proposedAlternative = vocabularyService.proposeAlternative("CC-BY-SA 4.0", items, new String[] {});

        assertThat(proposedAlternative).hasValue("CC BY-SA 4.0");
    }

    @ParameterizedTest
    @Timeout(60000)
    @CsvSource(
        {
            "CC-BY-SA 4.0, Did you mean 'CC BY-SA 4.0' instead of 'CC-BY-SA 4.0'?",
            "CC-BY-SA-4.0, Did you mean 'CC BY-SA 4.0' instead of 'CC-BY-SA-4.0'?",
            "CC-BY-SA-4;0, Did you mean 'CC BY-SA 4.0' instead of 'CC-BY-SA-4;0'?",
            "CC-BY-SA-3.0, Did you mean 'CC BY-SA 3.0' instead of 'CC-BY-SA-3.0'?",
            "Weit Weg, ''",
        }
    )
    void testAlternativesProposer(String input, String output) {
        AlternativesProposer ap = vocabularyService.createAlternativesProposer(input, "License");
        assertThat(ap).hasToString(output);
    }

    @Test
    @Timeout(60000)
    void testFieldCheck() {
        ValidationResultDTO result1 = new ValidationResultDTO();
        vocabularyService.checkField(new String[] {}, "keyword", "Keyword", true, false, new String[] {}, result1);
        org.hamcrest.MatcherAssert.assertThat(
            "Should contain mandatory warning",
            result1.errors,
            hasItem(is("The field keyword is mandatory"))
        );

        UserProvidedMetadataDTO metaData = new UserProvidedMetadataDTO();
        metaData.setKeyword(new String[] { "Variables", "VariablenX", "VariablenX2", "VariablenX3", "Ganz was anderes" });

        ValidationResultDTO result2 = new ValidationResultDTO();
        vocabularyService.checkField(metaData.getKeyword(), "keyword", "Keyword", true, false, new String[] { "VariablenX" }, result2);
        org.hamcrest.MatcherAssert.assertThat(
            "Should contain value warning",
            result2.warnings,
            containsInAnyOrder(
                "The field keyword contains an unknown entry 'VariablenX2'. Did you mean 'Variablen' instead of 'VariablenX2'?",
                "The field keyword contains an unknown entry 'VariablenX3'. Did you mean 'Variablen' instead of 'VariablenX3'?",
                "The field keyword contains an unknown entry 'Ganz was anderes'."
            )
        );
    }

    @Test
    @Timeout(60000)
    void testReadConfig() throws StreamReadException, DatabindException, IOException {
        final VocabularyServiceConfig config = vocabularyService.readVocabularyConfig();
        for (VocabularySetting setting : config.getConfig()) {
            Assert.assertNotNull("getter for " + setting.getProperty() + " must not be null", setting.getGetter());
        }
    }

    @Test
    @Timeout(60000)
    void testDistance() {
        LanguageItem liDE = new LanguageItem();
        liDE.setTitle("Variablen");
        LanguageItem liEN = new LanguageItem();
        liEN.setTitle("Variables");
        VocabularyItem vi1 = new VocabularyItem();
        vi1.setLanguageItem(new LanguageItem[] { liDE });

        float dist = vocabularyService.getLevenSteinDistance(vi1, "Variablen");
        assertThat(dist).isEqualTo(1.0f, Offset.offset(0.001f));

        dist = vocabularyService.getLevenSteinDistance(vi1, "Variables");
        assertThat(dist).isEqualTo(0.8888f, Offset.offset(0.01f));

        VocabularyItem viDEEN = new VocabularyItem();
        viDEEN.setLanguageItem(new LanguageItem[] { liDE, liEN });

        dist = vocabularyService.getLevenSteinDistance(viDEEN, "Variablen");
        assertThat(dist).isEqualTo(1.0f, Offset.offset(0.001f));

        dist = vocabularyService.getLevenSteinDistance(viDEEN, "Variables");
        assertThat(dist).isEqualTo(1.0f, Offset.offset(0.01f));

        Optional<String> proposedAlternative = vocabularyService.proposeAlternative(
            "Variablen",
            new VocabularyItem[] { viDEEN },
            new String[] {}
        );
        assertThat(proposedAlternative).hasValue("Variablen");

        proposedAlternative = vocabularyService.proposeAlternative("Variables", new VocabularyItem[] { viDEEN }, new String[] {});
        assertThat(proposedAlternative).hasValue("Variables");
    }

    @ParameterizedTest
    @MethodSource("getAllVocabularies")
    @Timeout(60000)
    void testObsoleteExtraEntries(String key, URI target) {
        //        if (key.equals("Requires") || key.equals("Teaches") || key.equals("Language") || key.equals("Contributors")
        //                || key.equals("Educational Level") || key.equals("Time Required") || key.equals("Creators"))
        //            return;
        VocabularyItem[] items = vocabularyRepository.getVocabularyItemsFor(key);
        final Optional<VocabularySetting> setting = vocabularyService.getConfig().getConfigSettingByOEResourceType(key);
        if (setting.isEmpty()) {
            // not relevant
            return;
        }

        final String[] extraEntries = setting.get().getExtraEntries();

        Arrays
            .stream(extraEntries)
            .forEach(extraEntry ->
                Assert.assertFalse(
                    String.format(
                        "Der Eintrag %s ist bereits in %s enthalten. Er kann in vocabularyServiceConfig.json gelöscht werden",
                        extraEntry,
                        key
                    ),
                    Arrays
                        .stream(items)
                        .anyMatch(item -> Arrays.stream(item.getLanguageItem()).anyMatch(li -> extraEntry.equals(li.getTitle())))
                )
            );
    }

    private Stream<Arguments> getVariousTestMetaData() {
        List<Arguments> testData = List.of(
            Arguments.of(
                "testMetaDataWrongDifficulty.yaml",
                List.of("The field difficulty contains an invalid entry 'advanced'."),
                Collections.emptyList()
            ),
            Arguments.of("testMetaDataCorrect.yaml", Collections.emptyList(), Collections.emptyList()),
            Arguments.of("testMetaDataWrongLanguage.yaml", List.of("The language code 'it' is not yet supported."), Collections.emptyList())
        );
        return testData.stream();
    }

    @ParameterizedTest
    @MethodSource("getVariousTestMetaData")
    @Timeout(60000)
    void testVariousMetaData(String fileName, List<String> expectedErrors, List<String> expectedWarnings)
        throws StreamReadException, DatabindException, IOException {
        ObjectMapper yamlMapper = new ObjectMapper(new YAMLFactory());
        yamlMapper.findAndRegisterModules();

        final InputStream testMetaDataStream = this.getClass().getResourceAsStream(fileName);
        ExtUserProvidedMetadataDTO testMetaData = yamlMapper.readValue(testMetaDataStream, ExtUserProvidedMetadataDTO.class);

        final ValidationResultDTO validateMetaData = vocabularyService.validateMetaData(testMetaData, true);

        assertThat(validateMetaData.getErrors()).containsExactlyInAnyOrderElementsOf(expectedErrors);
        assertThat(validateMetaData.getWarnings()).containsExactlyInAnyOrderElementsOf(expectedWarnings);
    }
}
